/* Modified by David Gesswein 4/10/01 */
#include "c42pdf.h"

long GetLong (FILE* tiffile,char byteorder)
{
long out=0;
if (byteorder == FILLORDER_LSB2MSB)
  { 
  out+=(long)getc(tiffile);
  out+=(long)getc(tiffile)*256;
  out+=(long)getc(tiffile)*256*256;
  out+=(long)getc(tiffile)*256*256*256;
  }
else
  {
  out+=(long)getc(tiffile)*256*256*256;
  out+=(long)getc(tiffile)*256*256;
  out+=(long)getc(tiffile)*256; 
  out+=(long)getc(tiffile);
  }
return out;
}

int GetInt (FILE* tiffile,char byteorder)
{
int out=0;
if (byteorder == FILLORDER_LSB2MSB)
  {
  out+=(int)getc(tiffile);
  out+=(int)getc(tiffile)*256;
  }
else
  {
  out+=(int)getc(tiffile)*256;
  out+=(int)getc(tiffile);  
  }
return out;
}

float resolutionfactor(TIFF *t) 
{
 switch (t->resolutionunit)
      {
      case RESUNIT_NONE: printf("Error, resolution unit in TIFF image specified as RESUNIT_NONE, can't get page size!\n"); exit(2); break;
      case RESUNIT_INCH: return 72.0; break;
      case RESUNIT_CENTIMETER: return 26.66667; break; 
      default: printf("Error, resolution unit in TIFF image is not defined: (%d)\n", t->resolutionunit); exit(2); break;
      }
}

int flippit (PDF *p, TIFF *t, options *o,int flip) {
        if (flip && (t->imagewidth*t->xresolution > t->imageheight*t->yresolution && o->paperwidth < o->paperheight || t->imageheight*t->yresolution > t->imagewidth*t->xresolution && o->paperheight < o->paperwidth))return true;  /*landscape orientation*/ 
	else
	  return false; 
}

void handle_fancy_options(PDF *p, TIFF *t, options *o)
{
        float buffer; 
	/*image size is resized to output format automatic paper resizing*/
         if ((o->autopapersize)) {
            if (t->xresolution * t->imagewidth > 11.1 || 
                t->yresolution * t->imageheight > 11.1) {
               o->paperwidth = p11x17.width;
               o->paperheight = p11x17.height;
            } else {
               o->paperwidth = letter.width;
               o->paperheight = letter.height;
            }
         } else
         if (!(o->fixedpapersize)) 
                {
		#ifdef DEBUG
			printf("handling fancy options"); 
		#endif
                o->paperwidth = t->xresolution * resolutionfactor(t) *t->imagewidth;
                o->paperheight = t->yresolution * resolutionfactor(t) * t->imageheight;
                }

	
                

}

int num_multi = 0;
void readtiffpage(PDF *p, TIFF *t, options *o)
{
  int entry_no; /*number of field entries in TIFF file*/
  int field_no; /*number of field analyzed*/
  int i;
  int nextstart = 0;

	fseek(t->tiffile,t->nextpageifdoffset,SEEK_SET); /*pointer to IFD*/
        entry_no = GetInt(t->tiffile,t->byteorder);
	for (i=0;i<entry_no;i++)
	  {
	  field_no = GetInt(t->tiffile,t->byteorder);
	 
	  fseek (t->tiffile,6,SEEK_CUR);
	  switch(field_no)
	    {
	    case TIFFTAG_IMAGEWIDTH: 
              t->imagewidth = (int)GetLong(t->tiffile,t->byteorder); break;
	    case TIFFTAG_IMAGELENGTH: 
              t->imageheight = (int)GetLong(t->tiffile,t->byteorder);break;
	    case TIFFTAG_STRIPHEIGHT: 
              t->stripheight = (int)GetLong(t->tiffile,t->byteorder);break;
	    case TIFFTAG_STRIPSIZES: 
             {
                 long loc,cntr;
	         fseek (t->tiffile,-4,SEEK_CUR);
                 t->num_strips = GetLong(t->tiffile,t->byteorder);
                 if (t->num_strips == 1) {
                    t->strip_sizes[0] = GetLong(t->tiffile,t->byteorder);
                 } else {
                       num_multi++;
                    t->strip_list_offset = GetLong(t->tiffile,t->byteorder);
                    loc = ftell(t->tiffile);
                    fseek(t->tiffile,t->strip_list_offset,SEEK_SET);
                    if (t->num_strips > sizeof(t->strip_offsets) / 
                        sizeof(t->strip_offsets[0])) {
                       printf("Too many strips %d\n",t->num_strips);
                       exit(1);
                    }
                    for (cntr = 0; cntr < t->num_strips; cntr++) {
                       t->strip_sizes[cntr] = GetLong(t->tiffile,t->byteorder);
                    }
                    fseek(t->tiffile,loc,SEEK_SET);
                 }
             }
            break;
	    case TIFFTAG_STRIPOFFSETS: 
              {
                 long loc,cntr;
	         fseek (t->tiffile,-4,SEEK_CUR);
                 t->num_strips = GetLong(t->tiffile,t->byteorder);
                 if (t->num_strips == 1) {
                    t->strip_offsets[0] = GetLong(t->tiffile,t->byteorder);
                 } else {
                    t->strip_list_offset = GetLong(t->tiffile,t->byteorder);
                    loc = ftell(t->tiffile);
                    fseek(t->tiffile,t->strip_list_offset,SEEK_SET);
                    if (t->num_strips > sizeof(t->strip_offsets) / 
                          sizeof(t->strip_offsets[0])) {
                       printf("Too many strips %d\n",t->num_strips);
                       exit(1);
                    }
                    for (cntr = 0; cntr < t->num_strips; cntr++) {
                       t->strip_offsets[cntr] = GetLong(t->tiffile,t->byteorder);
                    }
                    fseek(t->tiffile,loc,SEEK_SET);
                 }
              }
              break;
	    case TIFFTAG_COMPRESSION: 
              t->compression = (int)GetLong(t->tiffile,t->byteorder); break;
            case TIFFTAG_FILLORDER: 
	      t->compressionfillorder = (char)GetLong(t->tiffile,t->byteorder); break; 
            case TIFFTAG_RESOLUTIONUNIT: 
	      t->resolutionunit = GetInt(t->tiffile,t->byteorder); fseek(t->tiffile,2,SEEK_CUR) ; break;
            case TIFFTAG_XRESOLUTION: 
            {
              long num, offset,loc;
              long v1,v2;

              fseek (t->tiffile,-4,SEEK_CUR);
              num = GetLong(t->tiffile,t->byteorder);
              offset = GetLong(t->tiffile,t->byteorder);
              if (num != 1) {
                 v1 = num;
                 v2 = offset;
              } else {
                 loc = ftell(t->tiffile);
                 fseek(t->tiffile,offset,SEEK_SET);
                 v2 = GetLong(t->tiffile,t->byteorder);
                 v1 = GetLong(t->tiffile,t->byteorder);
                 fseek(t->tiffile,loc,SEEK_SET);
              }

              t->xresolution = (float) v1/v2;
            }
            break;
            case TIFFTAG_YRESOLUTION: 
            {
              long num, offset,loc;
              long v1,v2;

              fseek (t->tiffile,-4,SEEK_CUR);
              num = GetLong(t->tiffile,t->byteorder);
              offset = GetLong(t->tiffile,t->byteorder);
              if (num != 1) {
                 v1 = num;
                 v2 = offset;
              } else {
                 loc = ftell(t->tiffile);
                 fseek(t->tiffile,offset,SEEK_SET);
                 v2 = GetLong(t->tiffile,t->byteorder);
                 v1 = GetLong(t->tiffile,t->byteorder);
                 fseek(t->tiffile,loc,SEEK_SET);
              }

              t->yresolution = (float) v1/v2;
            }
            break;
	    default: 
              fseek(t->tiffile,4,SEEK_CUR);
	    }
	  }
        /*pointer to next fdi*/
        (o->runs)++;
        /*t->thispagedataend = GetLong(t->tiffile,t->byteorder);*/
        nextstart = GetLong(t->tiffile,t->byteorder);
	if (o->runs == o->repetitions) 
	         {
                 t->nextpageifdoffset = nextstart;
                 o->runs = 0;
                 } 


	handle_fancy_options(p,t,o);
	
	p->width=o->paperwidth; p->height=o->paperheight; /*portrait orientation*/
    
        if (flippit(p,t,o,!o->flip)) 
	  {float buffer; 
		#ifdef DEBUG
		  printf ("flipping "); 
		#endif
		  buffer=p->width;
		  p->width=p->height; 
		  p->height = buffer; 
	  }  
        /*landscape orientation*/      
         

	#ifdef DEBUG
        printf("compression %d strips %ld strip offset %ld nextpageifdoffset %ld height %d width %d entryno %d xresolution %f yresolution %f resolutionunit %d papaerwidth %f paperheight %f pagewidth %f pageheight %f\n",t->compression,t->num_strips,t->strip_list_offset,t->nextpageifdoffset,t->imageheight,t->imagewidth,entry_no,t->xresolution,t->yresolution,t->resolutionunit,o->paperwidth,o->paperheight, p->width, p->height);
	#endif
        if (t->compression!=COMPRESSION_CCITTFAX4) {printf("No CCITT4 compression in TIFF source file (tag for compression type=%d). c42pdf cannot handle this file !", t->compression); exit(2);}	
}

void readtifffile(PDF * p, TIFF *t, options * o)
{
  int cntr;

  num_multi = 0;

  /*defaults*/
  t->compression = 0; 
  t->compressionfillorder = FILLORDER_MSB2LSB; /*(if tag 266 is missing)*/
  t->pagecount = 0;
  t->resolutionunit = RESUNIT_NONE;
  t->xresolution = 1.0; t->yresolution = 1.0; 

  t->byteorder = getc(t->tiffile);
    switch(t->byteorder)
      {
      case 'I': t->byteorder = FILLORDER_LSB2MSB; break; /*Intel little endian*/
      case 'M': t->byteorder = FILLORDER_MSB2LSB; break; /*Motorola big endian*/
      default: printf("Error: This is not a TIFF file (it ought to begin with I or M when viewed in ASCII editor and not with \"%c\").\n", t->byteorder); exit(2);
      }
    fseek(t->tiffile,4,SEEK_SET);
    t->nextpageifdoffset = GetLong(t->tiffile,t->byteorder);
        /*position of first IFD pointer in header*/
    
    do
      { 
	float widthbuffer, heightbuffer; 
	readtiffpage(p,t,o);

        if (t->pagecount >= o->first_page-1 && t->pagecount <= o->last_page-1) {
	   widthbuffer = p->width; heightbuffer = p->height; 
           PDF_begin_page(p, p->width, p->height);
	   p->height = heightbuffer; p->width = widthbuffer; 
 
           /* define outline with filename */

           for (cntr = 0; cntr < t->num_strips; cntr++) {
              C4_put_image(p, t, o, cntr);
              C4_execute_image(p, t, o, cntr);
           }

	   p->res.procset = ImageB;
           C4_PDF_end_page(p,o->crop,o->runs);
        }

        t->pagecount++;
      } while (t->nextpageifdoffset);
    if (fclose(t->tiffile)!=0) {printf ("Error in closing file."); exit(1);}
    if (!o->quiet && num_multi != 0) 
       printf("%d of %d pages multistrip\n",num_multi,t->pagecount);
}













